
import numpy as np
import pandas as pd
from pandas.io.formats.style import Styler
from xmle import Elem
import matplotlib.figure, matplotlib.axes
import io
import time
import base64
from .png import make_png
from .. import __version__
import logging

logger = logging.getLogger("Larch.Excel")

try:
    import xlsxwriter
except ImportError:
    xlsxwriter = None
    _XlsxWriter = object
else:
    try:
        from pandas.io.excel._xlsxwriter import _XlsxWriter
    except ImportError:
        from pandas.io.excel import _XlsxWriter


class NumberedCaptions:

    def __init__(self, kind):
        self._kind = kind
        self._n = 1

    def __call__(self, caption):
        if caption:
            numb_caption = f"{self._kind} {self._n}: {caption}"
            self._n += 1
            return numb_caption
        else:
            return caption

class ExcelWriter(_XlsxWriter):
    engine = 'xlsxwriter_larch'
    supported_extensions = ('.xlsx',)

    def __init__(
            self,
            *args,
            model=None,
            data_statistics=True,
            nesting=True,
            numbering=True,
            hide_log=True,
            output_renderer=None,
            embed_model=True,
            **kwargs,
    ):

        _engine = 'xlsxwriter_larch'
        kwargs.pop('engine', None)
        options = kwargs.pop('workbook_options', {})
        if 'nan_inf_to_errors' not in options:
            options['nan_inf_to_errors'] = True
        super().__init__(*args, engine=_engine, options=options, **kwargs)
        self.book.set_size(1600, 1200)
        self.head_fmt = self.book.add_format({'bold': True, 'font_size':14})
        self.ital_fmt = self.book.add_format({'italic': True, 'font_size':12})
        self.toc_link_fmt = self.book.add_format({'font_size':8})

        self.fixed_precision = {
            2: "_-0.00;-0.00"
        }

        self.sheet_startrow = {}
        self._col_widths = {}
        if output_renderer is None:
            output_renderer = lambda x: x
        self._output_renderer = output_renderer

        if numbering:
            self.TAB = NumberedCaptions('Table')
            self.FIG = NumberedCaptions('Figure')
        else:
            self.TAB = lambda x: x
            self.FIG = lambda x: x

        self.tocsheet = self.add_worksheet('Contents') # first sheet cannot be hidden
        if model is not None:
            self.tocsheet.write(0, 0, model.title, self.head_fmt)
        self.tocsheet.write(2, 0, f'   🌳 Larch {__version__}')
        self.tocsheet.write(3, 0, time.strftime("     %A %d %B %Y, %I:%M:%S %p %Z"))
        self.tocsheet.write(5, 0, 'Table of Contents', self.head_fmt)
        self.sheet_startrow['Contents'] = 6

        self.logsheet = self.add_worksheet('_log_', hide=hide_log)
        self.log(f"larch.util.excel.ExcelWriter opened: {str(args)}")

        self.metadatasheet = self.add_worksheet('_metadata_', hide=True)
        self.metadatasheet.write(0, 0, 'key')
        self.metadatasheet.write(0, 1, 'len')
        self.metadatasheet.write(0, 2, 'val')
        self.sheet_startrow['_metadata_'] = 1

        if model is not None:
            self.add_model(model, data_statistics=data_statistics, nesting=nesting, embed=embed_model)


    def add_model(
            self,
            model,
            data_statistics=True,
            nesting=True,
            utility_functions=True,
            on_error='pass',
            embed=True,
    ):

        ps_data_index_nlevels = 1
        try:
            ps = model.parameter_summary('df')
            ps_data_index_nlevels = ps.data.index.nlevels
            if 'Constrained' in ps.data.columns:
                ps.data['Constrained'] = ps.data['Constrained'].str.replace("<br>",",\n")
            self.add_content_tab(ps, sheetname="Parameters", heading="Parameters" )
            if 'Constrained' in ps.data.columns:
                n = ps.data.index.nlevels + ps.data.columns.get_loc("Constrained")
                if "Parameters" not in self._col_widths:
                    self._col_widths["Parameters"] = {}
                current_width = self._col_widths["Parameters"].get(n, 8)
                new_width = max(current_width, 50)
                self._col_widths["Parameters"][n] = new_width
                wrap = self.book.add_format()
                wrap.set_text_wrap()
                self.sheets["Parameters"].set_column(n, n, width=new_width, cell_format=wrap)
        except:
            if on_error == 'raise':
                raise

        try:
            startrow = self.sheet_startrow.get("Parameters", 0)
            worksheet = self.sheets["Parameters"]
            worksheet.write_url(startrow, 0, 'internal:Contents!A1', self.toc_link_fmt, string='<< Back to Table of Contents', )
            startrow += 1
            h = self.TAB("Estimation Statistics")
            worksheet.write(startrow, 0, h, self.head_fmt)
            self._add_toc_entry(h, "Parameters", startrow)
            startrow += 1
            logger.debug(" after heading row is %d", startrow)
            model._estimation_statistics_excel(self, "Parameters", start_row=startrow, buffer_cols=ps_data_index_nlevels-1)
        except:
            if on_error == 'raise':
                raise

        if data_statistics:
            from .statistics import statistics_for_dataframe
            if model.dataframes is not None:
                if model.dataframes.data_co is not None and len(model.dataframes.data_co.columns):
                    try:
                        self.add_content_tab(
                            statistics_for_dataframe(model.dataframes.data_co),
                            sheetname="CO Data",
                            heading="CO Data",
                        )
                    except:
                        if on_error=='raise':
                            raise
                if model.dataframes.data_ca is not None and len(model.dataframes.data_ca.columns):
                    try:
                        self.add_content_tab(
                            statistics_for_dataframe(model.dataframes.data_ca),
                            sheetname="CA Data",
                            heading="CA Data",
                        )
                    except:
                        if on_error=='raise':
                            raise
                if model.dataframes.data_ce is not None and len(model.dataframes.data_ce.columns):
                    try:
                        self.add_content_tab(
                            statistics_for_dataframe(model.dataframes.data_ce),
                            sheetname="CE Data",
                            heading="CE Data",
                        )
                    except:
                        if on_error=='raise':
                            raise
                if model.dataframes.data_ch is not None and model.dataframes.data_av is not None:
                    try:
                        self.add_content_tab(
                            model.dataframes.choice_avail_summary(
                                graph=model.graph,
                                availability_co_vars=model.availability_co_vars,
                            ),
                            sheetname="Choice",
                            heading="Choices",
                        )
                    except:
                        if on_error=='raise':
                            raise

        if utility_functions:
            try:
                self.add_content_tab(model._utility_functions_as_frame(), sheetname="Utility", heading="Utility Functions")
                self.sheets["Utility"].set_column('B:B', None, None, {'hidden': 1})
            except:
                if on_error == 'raise':
                    raise

        if nesting and not model.is_mnl():
            try:
                self.add_content_tab(model.graph.__xml__(output='png'), sheetname="Nesting", heading="Nesting Tree")
            except:
                if on_error == 'raise':
                    raise
            try:
                self.add_content_tab(model.graph_descrip('nodes'), sheetname="Nesting", heading="Nesting Node List")
            except:
                if on_error == 'raise':
                    raise

        if embed:
            self.add_metadata('_self_', model)

    def add_metadata(self, key, value):
        if not isinstance(key, str):
            raise ValueError(f"metadata key must be str not {type(key)}")
        import base64
        try:
            import cloudpickle as pickle
        except ImportError:
            import pickle
        encoded_value = base64.standard_b64encode(pickle.dumps(value)).decode()
        row = self.sheet_startrow.get('_metadata_',0)
        self.metadatasheet.write(row, 0, key)
        c = 2
        while len(encoded_value):
            self.metadatasheet.write(row, c, encoded_value[:30_000]) # excel has cell text length cap
            c += 1
            encoded_value = encoded_value[30_000:]
        self.metadatasheet.write(row, 1, c-2)
        self.sheet_startrow['_metadata_'] = row+1

    def log(self, message):
        t = time.strftime("%Y-%b-%d %H:%M:%S")
        row = self.sheet_startrow.get('_log_',0)
        self.logsheet.write(row, 0, t)
        self.logsheet.write(row, 1, str(message))
        self.sheet_startrow['_log_'] = row+1

    def add_worksheet(self, name, force=False, hide=None):

        if not force and name in self.sheets:
            return self.sheets[name]

        s = None
        try:
            s = self.book.add_worksheet(name)
        except xlsxwriter.exceptions.DuplicateWorksheetName:
            i = 2
            while s is None:
                try:
                    s = self.book.add_worksheet(f"{name}{i}")
                except xlsxwriter.exceptions.DuplicateWorksheetName:
                    i += 1
                    if i > 99:
                        raise
        self.sheets[name] = s
        if hide is not None:
            if hide:
                s.hide()
            else:
                s.hidden = 0

        return s

    def _add_toc_entry(self, heading, target_sheet, target_row):
        self.tocsheet.write_url(self.sheet_startrow['Contents'], 0, f"internal:'{target_sheet}'!A{target_row+1}", string=heading)
        self.sheet_startrow['Contents'] += 1

    def add_content_tab(self, content, sheetname, heading=None, startrow=None, blurb=None, to_excel=None):

        if startrow is None:
            startrow = self.sheet_startrow.get(sheetname, 0)
        worksheet = self.add_worksheet(sheetname)

        content_in = content
        success = False
        if to_excel is None:
            to_excel = {}

        logger.debug("writing to %s", sheetname)

        if isinstance(content, (pd.DataFrame, Styler)):
            logger.debug("writing as DataFrame (%s)", type(content))

            if isinstance(content, Styler):
                content_, content_data = content, content.data
            else:
                content_, content_data = content, content

            if heading is not None:
                worksheet.write_url(startrow, 0, 'internal:Contents!A1', self.toc_link_fmt, string='<< Back to Table of Contents', )
                startrow += 1
                h = self.TAB(heading)
                worksheet.write(startrow, 0, h, self.head_fmt)
                self._add_toc_entry(h, sheetname, startrow)
                startrow += 1
                logger.debug(" after heading row is %d", startrow)
            content_.to_excel(self, sheet_name=worksheet.name, startrow=startrow, **to_excel)
            startrow += len(content_data) + content.columns.nlevels
            if content_data.index.name:
                startrow += 1
            startrow += 2 # gap
            logger.debug(" after table row is %d", startrow)
            success = True

            if sheetname not in self._col_widths:
                self._col_widths[sheetname] = {}
            if content_data.index.nlevels == 1:
                current_width = self._col_widths[sheetname].get(0, 8)
                new_width = max(current_width, max(len(str(i)) for i in content_data.index))
                self._col_widths[sheetname][0] = new_width
                worksheet.set_column(0, 0, new_width)
            else:
                for n in range(content_data.index.nlevels):
                    current_width = self._col_widths[sheetname].get(n, 8)
                    new_width = max(current_width, max(len(str(i)) for i in content_data.index.levels[n]))
                    self._col_widths[sheetname][n] = new_width
                    worksheet.set_column(n,n,new_width)

        # Render matplotlib.Figure into an Elem
        if not success and 'matplotlib' in str(type(content)):
            logger.debug("writing as matplotlib (%s)", type(content))
            if isinstance(content, matplotlib.axes.Axes):
                content = content.get_figure()
            if isinstance(content, matplotlib.figure.Figure):
                try:
                    content = make_png(content, dpi='figure', compress=True, output='Elem', facecolor='w')
                except:
                    pass

        # Extract PNG data from Elem if found there
        if not success and isinstance(content, Elem):
            logger.debug("writing as Elem (%s)", type(content))
            if content.tag == 'img':
                _v = content
            else:
                _v = content.find('img')
            if _v is not None:
                try:
                    dpi = _v.attrib['dpi']
                except:
                    dpi = None

                try:
                    _v = _v.attrib['src']
                except:
                    pass
                else:
                    if isinstance(_v, str) and _v[:22] == 'data:image/png;base64,':
                        fp, img_size, dpi = make_png(
                            base64.standard_b64decode(_v[22:]), dpi=dpi, output='bytesio',
                            return_size=True, return_dpi=True,
                        )
                        if heading is not None:
                            worksheet.write_url(startrow, 0, 'internal:Contents!A1', self.toc_link_fmt,
                                                string='<< Back to Table of Contents', )
                            startrow += 1
                            h = self.FIG(heading)
                            worksheet.write(startrow, 0, h, self.head_fmt)
                            self._add_toc_entry(h, sheetname, startrow)
                            startrow += 1
                            logger.debug(" after heading row is %d", startrow)
                        worksheet.insert_image(startrow, 0, f'image-{sheetname}.png', {'image_data': fp})
                        startrow += int(np.ceil(img_size[1] / dpi[1] * 72 / 15)) + 3
                        logger.debug(" after image row is %d", startrow)
                        success = True

        if not success and isinstance(content, dict):
            logger.debug("writing as dict (%s)", type(content))
            max_row = 0
            max_col = 0
            row = -1
            prev_col = 0
            cellcontent = []
            from collections.abc import Iterable
            def unpacker(i, indent=0):
                nonlocal row, prev_col, max_row, max_col, cellcontent
                if isinstance(i, dict):
                    for k, v in i.items():
                        unpacker(k, indent)
                        unpacker(v, indent + 1)
                elif isinstance(i, pd.Series):
                    for k, v in dict(i).items():
                        unpacker(k, indent)
                        unpacker(v, indent + 1)
                elif not isinstance(i, str) and isinstance(i, Iterable):
                    for k in i:
                        unpacker(k, indent)
                else:
                    if prev_col >= indent:
                        row = row + 1
                    cellcontent.append((indent, row, i))
                    max_col = max(max_col, indent)
                    max_row = max(max_row, row)
                    prev_col = indent
            unpacker(content)
            table_cache = pd.DataFrame(data="", index=pd.RangeIndex(max_row + 1), columns=pd.RangeIndex(max_col + 1))
            for c in cellcontent:
                table_cache.iloc[c[1], c[0]] = c[2]
            self.add_content_tab(
                table_cache,
                sheetname=sheetname,
                heading=heading,
                startrow=startrow,
                blurb=blurb,
                to_excel=dict(
                    index=False,
                    header=False,
                ),
            )
            if sheetname not in self._col_widths:
                self._col_widths[sheetname] = {}
            for n in range(max_col):
                current_width = self._col_widths[sheetname].get(n, 8)
                new_width = max(current_width, 16)
                self._col_widths[sheetname][n] = new_width
                worksheet.set_column(n,n,new_width)
            success = True

        if not success:
            import warnings
            warnings.warn(f'content not written to sheet {sheetname}, type {type(content_in)}')
        else:
            if blurb is not None:
                import textwrap
                blurb = textwrap.wrap(blurb)
                for line in blurb:
                    worksheet.write(startrow, 0, line)
                    startrow += 1
                logger.debug(" after blurb row is %d", startrow)

        self.sheet_startrow[worksheet.name] = startrow
        return self._output_renderer(content_in)

    def save(self, makedirs=True, overwrite=False):
        if self.path is not None:
            if makedirs:
                import os
                dirname = os.path.dirname(self.path)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)

            if not overwrite and not getattr(self, '__file_archived', False): # don't move twice
                from xmle.file_util import archive_existing_file
                try:
                    new_name = archive_existing_file(self.path, archive_path=None, tag='creation')
                except FileNotFoundError:
                    pass
                else:
                    self.log(f"archived existing file to {new_name}")
                    setattr(self, '__file_archived', new_name)

        self.log(f"saving")
        super().save()

if xlsxwriter is not None:
    from pandas.io.excel import register_writer
    register_writer(ExcelWriter)

def _make_excel_writer(model, filename, save_now=True, **kwargs):
    """
    Write the model to an Excel file.

    Parameters
    ----------
    model : larch.Model
    filename : str
    save_now : bool, default True
        Save the model immediately.  Set to False if you want to
        write additional figures or tables to the file before saving.
    **kwargs
        Other keyword arguments are passed to the `ExcelWriter`
        constructor.

    Returns
    -------
    larch.util.excel.ExcelWriter
    """
    if xlsxwriter is None:
        raise RuntimeError('xlsxwriter is not installed')
    xl = ExcelWriter(filename, engine='xlsxwriter_larch', model=model, **kwargs)
    if save_now:
        xl.save()
    return xl

from .. import Model
Model.to_xlsx = _make_excel_writer

def load_metadata(xlsx_filename, key=None):
    import pickle, base64, pandas
    raw = pandas.read_excel(xlsx_filename, sheet_name='_metadata_', index_col=0)
    metadata = {}
    if key is None:
        for key, row in raw.iterrows():
            v = (row[1:row[0] + 1].str.cat())
            metadata[key] = pickle.loads(base64.standard_b64decode(v.encode()))
        return metadata
    else:
        row = raw.loc[key]
        v = (row[1:row[0] + 1].str.cat())
        return pickle.loads(base64.standard_b64decode(v.encode()))




from .h5data import *
from .h5vault import *

import tables
import sys, atexit
from distutils.version import StrictVersion

def my_close_open_files(verbose):
    open_files = tables.file._open_files

    are_open_files = len(open_files) > 0

    if verbose and are_open_files:
        sys.stderr.write("Closing remaining open files:")

    if StrictVersion(tables.__version__) >= StrictVersion("3.1.0"):
        # make a copy of the open_files.handlers container for the iteration
        handlers = list(open_files.handlers)
    else:
        # for older versions of pytables, setup the handlers list from the
        # keys
        keys = open_files.keys()
        handlers = []
        for key in keys:
            handlers.append(open_files[key])

    for fileh in handlers:
        if verbose:
            sys.stderr.write("\n%s..." % fileh.filename)

        fileh.close()

        if verbose:
            sys.stderr.write("done")

    if verbose and are_open_files:
        sys.stderr.write("\n")

atexit.register(my_close_open_files, False)



import io
import base64
from xmle import Elem
from PIL import Image


def make_png(
		content,
		dpi=None,
		compress=True,
		output='Elem',
		close_after=True,
		pad_inches=0.1,
		facecolor=None,
		return_size=False,
		return_dpi=False,
):

	_bytes_io = None
	_img = None
	_size = None

	if hasattr(content, '_repr_png_'):
		content = content._repr_png_()

	if isinstance(content, bytes):
		_bytes_io = io.BytesIO(content)

	if 'matplotlib' in str(type(content)):
		from matplotlib import pyplot as plt
		import matplotlib.figure
		if not isinstance(content, matplotlib.figure.Figure):
			try:
				content = content.get_figure()
			except AttributeError:
				if not hasattr(content, 'savefig'):
					raise TypeError('matplotlib content must provide `get_figure` or `savefig` method.')

		# content is a Figure or otherwise has a savefig method
		try:
			fig_number = content.number
		except AttributeError:
			fig_number = None

		if facecolor is None:
			try:
				facecolor = content.get_facecolor()
			except:
				facecolor = 'w'

		try:
			edgecolor = content.get_edgecolor()
		except:
			edgecolor = 'none'

		_bytes_io = io.BytesIO()
		content.savefig(
			_bytes_io,
			dpi=dpi,
			orientation='portrait',
			format='png',
			bbox_inches='tight',
			pad_inches=pad_inches,
			facecolor=facecolor,
			edgecolor=edgecolor,
		)

		if close_after and fig_number is not None:
			plt.close(fig_number)

	if _bytes_io.getvalue()[:4] == b'\x89PNG':
		_img = Image.open(_bytes_io)
		dpi = _img.info.get('dpi', dpi)
		if dpi is None:
			dpi = [96, 96]
		if compress:
			_img = _img.convert(mode='P', palette='ADAPTIVE')
		_bytes_io = io.BytesIO()
		_img.save(_bytes_io, dpi=dpi, format='png')
		_size = _img.size
	else:
		raise ValueError("Not valid PNG data")

	if output.lower() == 'elem':
		result = Elem(
			tag='img',
			src="data:image/png;base64,{}".format(base64.standard_b64encode(_bytes_io.getvalue()).decode()),
		)
	elif output.lower() == 'bytesio':
		result = _bytes_io
	else:
		result = _img

	if return_size:
		if isinstance(result, tuple):
			result = (*result, _size)
		else:
			result = result, _size

	if return_dpi:
		if isinstance(result, tuple):
			result = (*result, dpi)
		else:
			result = result, dpi

	return result




from larch.examples import example
from larch import Model
from larch.roles import PX
from pytest import approx

def test_301():

	d = example(300, 'd')

	m = Model(dataservice=d)

	v = [
		"timeperiod==2",
		"timeperiod==3",
		"timeperiod==4",
		"timeperiod==5",
		"timeperiod==6",
		"timeperiod==7",
		"timeperiod==8",
		"timeperiod==9",
		"carrier==2",
		"carrier==3",
		"carrier==4",
		"carrier==5",
		"equipment==2",
		"fare_hy",
		"fare_ly",
		"elapsed_time",
		"nb_cnxs",
	]

	m.utility_ca = sum(PX(i) for i in v)
	m.choice_ca_var = 'choice'
	m.load_data()

	result = m.maximize_loglike()

	assert result.loglike == approx(-777770.0688722526)
	assert result.x['carrier==2'] == approx(0.11720047917232307)
	assert result.logloss == approx(3.306873650593341)




from .generic import *
import warnings

class H5PodCE(H5Pod):
	"""
	A HDF5 group node containing :ref:`idce` format data.

	The CE pod is conceptually similar to the CA pod, although alternatives are included
	element-wise, so that each array of data is stored as a vector, and casewise boundaries
	are permitted to be ragged.

	The H5PodCE has two special vectors, indicating the case-index and alt-index position of
	each item within the array, respectively.

	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._caseindex = None
		self._altindex = None

		# to translate ids to indexes...
	    # uniq_labels, uniq_indexes = numpy.unique(labels, return_inverse=True)

	@property
	def podtype(self):
		return 'idce'

	@property
	def n_cases(self):
		if self._caseindex is None:
			raise ValueError("no caseindexes set")
		return self._groupnode._v_children[self._caseindex][:].max()+1

	@property
	def n_alts(self):
		if self._altindex is None:
			raise ValueError("no altindexes set")
		return self._groupnode._v_children[self._altindex][:].max()+1

	@property
	def shape(self):
		"""The shape of the represented idca data (i.e., cases by alts)."""
		try:
			return (self.n_cases, self.n_alts, )
		except ValueError:
			raise NoKnownShape()

	@property
	def metashape(self):
		"""The shape of the underlying raw data (i.e., number of listed casealts)."""
		return super().shape


	def create_indexes_from_labels(self, indexes_name, labels_name=None, labels_data=None, make=None, **kwargs):
		"""

		Parameters
		----------
		indexes_name : str
			Name for the new array to be created.
		labels_name : str, optional
			Name of the existing array that gives the labels to be uniquely indexed.
		labels_data : ndarray, optional
			Array of labels.  Must have same shape as self.
		make : {'cases','alts',None}
			The newly created indexes will become this dimension for the outputs.
		kwargs
			passed to add_array

		Returns
		-------
		uniq_labels : ndarray
			An ordered array of labels corresponding to the indexes created.
		"""

		if labels_name is None and labels_data is None:
			raise ValueError("one of labels_name or labels_data must be given")

		if make not in ('cases','alts',None):
			raise ValueError("make must be in {'cases','alts',None}")

		if labels_data is None:
			labels_data = self._groupnode._v_children[labels_name][:]

		try:
			if labels_data.shape != self.metashape:
				raise IncompatibleShape(f'labels_data.shape of {labels_data.shape} does not match self.metashape {self.metashape}')
		except NoKnownShape:
			pass

		uniq_labels, uniq_indexes = numpy.unique(labels_data, return_inverse=True)

		self.add_array(indexes_name, uniq_indexes, **kwargs)

		if make == 'cases':
			self._caseindex = indexes_name
		elif make == 'alts':
			self._altindex = indexes_name

		return uniq_labels


	def set_casealt_indexes(self, case_index, alts_index):
		self._caseindex = case_index
		self._altindex = alts_index


	def merge_external_data(
			self, other,
			self_on, other_on=None, other_on_index=False,
			dupe_suffix="_copy", original_source=None, names=None, overwrite=False,
			other_dedupe=False, log=lambda *x: None, **kwargs
	):
		"""
		Merge data into this H5PodCE.

		Every casealt in the current (receiving) H5PodCO should match one or zero casealts in the
		imported data.

		Parameters
		----------
		other : H5PodCO or pandas.DataFrame or str
			The other data table to be merged.  Can be another DT, or a DataFrame, or the
			path to a file of type {csv, xlsx, dbf}.
		self_on : label or list, or array-like
			Field names to join on in this H5PodCE. Can be a vector or list of vectors
			of correct length to use a particular vector as the join key(s) instead of columns
		other_on : label or list, or array-like
			Field names to join on in other DataFrame, or vector/list of vectors per self_on
		dupe_suffix : str
			A suffix to add to variables that duplicate variables names already in this H5PodCE
		original_source : str, optional
			Give the original source of this data.  If not given and the filename can be
			inferred from other, that name will be used.
		names : iterable or dict, optional
			If given as a list, only these names will be merged from the other data.
			If given as a dict, the keys are the names that will be merged and the
			values are the new names in this DT.
		"""
		if other_on is None:
			if not other_on_index:
				other_on = self_on
			else:
				other_on = []

		if names is not None and not isinstance(names, dict):
			if isinstance(names, str):
				names = {names: names}
			else:
				names = {n: n for n in names}

		def _warn_nothing_done_():
			if names is not None:
				names_warn = '\n  from names:\n    ' + '\n    '.join(names.keys())
			else:
				names_warn = ''
			warnings.warn('nothing was imported into {}{}'.format(self.filename, names_warn))

		if isinstance(self_on, str):
			self_on = (self_on,)
		if isinstance(other_on, str):
			other_on = (other_on,)

		if self_on is None:
			baseframe = None
		elif isinstance(self_on, str):
			baseframe = self.dataframe[self_on]
		else:
			baseframe = self.dataframe[tuple(self_on)]
		anything_imported = False

		# get the other data (to be merged) as a pandas DataFrame
		if isinstance(other, H5PodCE):
			# we have a H5PodCO
			og = other._groupnode
			if other_on_index:
				raise TypeError('other_on_index cannot be True for other as H5PodCO')
			if original_source is None:
				original_source = other._h5f.filename
			if names is None:
				# we are potentially importing everything, so load everything into memory
				other_df = other.dataframe[other.names()]
			else:
				# we are importing only some things, so load only what will be needed
				use_names = set(names.keys()) | (set(other_on))
				for col in other.names():
					if col in other_on:
						continue
					if col not in names or (names[col] in self._groupnode and not overwrite):
						use_names.discard(col)
				if use_names == set(other_on):
					# there is nothing left to load but the matching keys,
					# so nothing will be loaded.  Abort now and save a bunch of work.
					_warn_nothing_done_()
					return
				other_df = other.dataframe[use_names]
		elif isinstance(other, pandas.DataFrame):
			# it is already a dataframe, so convenient!
			other_df = other
			og = None
		else:
			raise TypeError("bad type for other data")

		if other_dedupe:
			try:
				other_df.drop_duplicates(
					subset=None if other_on_index else other_on,
					keep='first',
					inplace=True
				)
			except TypeError as err:
				raise ValueError("Drop Duplicates Error") from err

		if baseframe is not None:
			new_df = pandas.merge(
				baseframe,
				other_df,
				left_on=self_on,
				right_on=None if other_on_index else other_on,
				right_index=other_on_index,
				how='left',
				suffixes=('', dupe_suffix),
			)
		else:
			new_df = other_df

		log(new_df)
		for col in new_df.columns:
			if names is None and col not in self._groupnode:
				log('importing "{}" into {}'.format(col, self.filename))
				try:
					self.add_array(col, arr=new_df[col].values)
				except UnicodeEncodeError:
					# The datatype has some problems converting to H5, warn and skip it.
					warnings.warn(f"unable to import {col}, the datatype is weird")
					continue
				anything_imported = True
				if original_source is not None:
					self._groupnode._v_children[col]._v_attrs.ORIGINAL_SOURCE = original_source
				if og is not None and col in og and 'DICTIONARY' in og._v_children[col]._v_attrs:
					self._groupnode._v_children[col]._v_attrs.DICTIONARY = og._v_children[col]._v_attrs.DICTIONARY
			elif names is not None and col in names and (names[col] not in self._groupnode or overwrite):
				log('importing "{}" as "{}" into {}'.format(col, names[col], self.filename))
				try:
					self.add_array(names[col], arr=new_df[col].values)
				except UnicodeEncodeError:
					# The datatype has some problems converting to H5, warn and skip it.
					warnings.warn(f"unable to import {col}, the datatype is weird")
					continue
				anything_imported = True
				if original_source is not None:
					self._groupnode._v_children[names[col]]._v_attrs.ORIGINAL_SOURCE = original_source
				if og is not None and col in og and 'DICTIONARY' in og._v_children[col]._v_attrs:
					self._groupnode._v_children[names[col]]._v_attrs.DICTIONARY = og._v_children[col]._v_attrs.DICTIONARY
		if not anything_imported:
			_warn_nothing_done_()


	@classmethod
	def from_dataframe(
			cls,
			df,
			h5filename=None,
			h5mode='a',
			h5groupnode='/',
			inmemory=False,
			temp=False,
			complevel=1,
			complib='zlib',
			ident=None,
			rename_columns=None,
			force_natural_names=True,
			caseindex=None,
			caselabels=None,
			altindex=None,
			altlabels=None,
	):
		self = cls(h5filename, h5mode, h5groupnode, inmemory=inmemory, temp=temp, complevel=complevel, complib=complib, ident=ident)
		if rename_columns:
			df.columns = [(rename_columns[c] if c in rename_columns else c) for c in df.columns]
		df.columns = [c.replace("/",'_') for c in df.columns]
		if force_natural_names:
			from ....util.naming import make_valid_identifier
			df.columns = [make_valid_identifier(c, True) for c in df.columns]
		self.merge_external_data(df, self_on=None, other_on=None)

		# if caseindex is None and caselabels is None:
		# 	raise ValueError("must give caseindex or caselabels")
		# if altindex is None and altlabels is None:
		# 	raise ValueError("must give caseindex or caselabels")

		if caseindex is not None and caseindex in self:
			self._caseindex = caseindex
			if caselabels is not None:
				import warnings
				warnings.warn("caselabels is ignored when caseindex is an existing column")
		else:
			if caseindex is None:
				import uuid
				caseindex = '_cIdx_'+str(uuid.uuid1()).replace('-','')
			if caselabels is not None:
				from ....warning import ignore_warnings
				with ignore_warnings():
					self.create_indexes_from_labels(caseindex, labels_name=caselabels)
				self._caseindex = caseindex

		if altindex is not None and altindex in self:
			self._altindex = altindex
			if altlabels is not None:
				import warnings
				warnings.warn("altlabels is ignored when altindex is an existing column")
		else:
			if altindex is None:
				import uuid
				altindex = '_aIdx_'+str(uuid.uuid1()).replace('-','')
			if altlabels is not None:
				from ....warning import ignore_warnings
				with ignore_warnings():
					self.create_indexes_from_labels(altindex, labels_name=altlabels)
				self._altindex = altindex

		return self

	@classmethod
	def from_csv(
			cls,
			*args,
			h5filename=None,
			h5mode='a',
			h5groupnode='/',
			inmemory=False,
			temp=False,
			complevel=1,
			complib='zlib',
			ident=None,
			rename_columns=None,
			force_natural_names=True,
			caseindex=None,
			caselabels=None,
			altindex=None,
			altlabels=None,
			**kwargs,
	):
		# self = cls(h5filename, h5mode, h5groupnode, inmemory=inmemory, temp=temp, complevel=complevel, complib=complib, ident=ident)
		df = pandas.read_csv(*args, **kwargs)
		return cls.from_dataframe(
			df,
			h5filename=h5filename,
			h5mode=h5mode,
			h5groupnode=h5groupnode,
			inmemory=inmemory,
			temp=temp,
			complevel=complevel,
			complib=complib,
			ident=ident,
			rename_columns=rename_columns,
			force_natural_names=force_natural_names,
			caseindex=caseindex,
			caselabels=caselabels,
			altindex=altindex,
			altlabels=altlabels,
		)

	@classmethod
	def from_excel(cls, *args, h5filename=None, h5mode='a', h5groupnode='/',
				   inmemory=False, temp=False, complevel=1, complib='zlib', ident=None,
				   rename_columns=None, force_natural_names=True,
				   **kwargs):
		self = cls(h5filename, h5mode, h5groupnode, inmemory=inmemory, temp=temp, complevel=complevel, complib=complib, ident=ident)
		df = pandas.read_excel(*args, **kwargs)
		if rename_columns:
			df.columns = [(rename_columns[c] if c in rename_columns else c) for c in df.columns]
		df.columns = [c.replace("/",'_') for c in df.columns]
		if force_natural_names:
			from ....util.naming import make_valid_identifier
			df.columns = [make_valid_identifier(c, True) for c in df.columns]
		self.merge_external_data(df, self_on=None, other_on=None)
		return self


	def as_array_idca(self, *names, caseindex=None, altindex=None, dtype=numpy.float64, present=None, **kwargs):

		if caseindex is None:
			caseindex = self._caseindex
		if altindex is None:
			altindex = self._altindex

		cx = self._groupnode._v_children[caseindex][:]
		ax = self._groupnode._v_children[altindex][:]

		shape = (cx.max()+1, ax.max()+1, len(names))
		result = numpy.zeros(shape, dtype=dtype)

		for nx,n in enumerate(names):
			if present == n:
				result[cx,ax,nx] = 1
			else:
				result[cx,ax,nx] = self._groupnode._v_children[n][:]

		return result


	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		return super().load_meta_data_item(name, result, selector=selector)

	def _shape_for_get_data_item(self, selector=None):
		if selector is not None:
			import warnings
			warnings.warn('selector not compatible for idce get_data_item')
		return self.metashape

	def new_alternative_codes(
			self,
			groupby,
			caseindex=None,
			name=None,
			create_index=None,
			padding_levels=4,
			groupby_prefixes=None,
			overwrite=False,
	):
		"""

		Parameters
		----------
		groupby
		caseindex
		name
		create_index
		padding_levels : int, optional
			Space for this number of "extra" levels is reserved in each mask, beyond the number of
			detected categories.  This is critical if these alternative codes will be used for OGEV models
			that require extra nodes at levels that are cross-nested.
		overwrite : bool, default False
			Should existing variables with `name` be overwritten. It False and they already exist, a
			NodeError is raised.

		Returns
		-------
		pandas.DataFrame, SystematicAlternatives

		"""

		if caseindex is None:
			caseindex = self._caseindex

		if isinstance(groupby, str):
			groupby = (groupby,)

		if groupby_prefixes is None:
			groupby_prefixes = ["" for g in groupby]

		df = self.dataframe.__getitem__( (caseindex, *groupby) )

		s = df.groupby([caseindex, *groupby]).cumcount()+1

		masks = numpy.zeros(len(groupby)+1, dtype=numpy.int64)

		label_offset = int(numpy.ceil(numpy.log2(s.max())))

		uniqs = [numpy.unique(df.iloc[:,i+1], return_inverse=True) for i,g in enumerate(groupby)]

		if len(groupby) > 0:
			n_uniqs = [len(uniqs[i][0])+1+padding_levels for i,g in enumerate(groupby)]
		else:
			n_uniqs = []

		bitmask_sizes = [int(numpy.ceil(numpy.log2(g))) for g in n_uniqs]

		#x = numpy.unique(df[groupby[0]], return_inverse=True)[1]+1
		if len(bitmask_sizes):
			x = uniqs[0][1]+1
			masks[0] = 2**bitmask_sizes[0]-1
		else:
			x = numpy.zeros(len(df), dtype=int)

		for i in range(1, len(bitmask_sizes)):
			x <<= bitmask_sizes[i]
			masks <<= bitmask_sizes[i]
			#x += numpy.unique(df[groupby[i]], return_inverse=True)[1]+1
			x += uniqs[i][1]+1
			masks[i] = 2**bitmask_sizes[i]-1

		x <<= label_offset
		masks <<= label_offset
		x += s
		masks[-1] = 2**label_offset-1

		if name is not None:
			self.add_array(name, x.values, overwrite=overwrite)

		if create_index is not None:
			altcodes = self.create_indexes_from_labels(
				indexes_name=create_index,
				labels_data=x,
				make='alts',
				overwrite=overwrite,
			)
		else:
			altcodes = numpy.unique(x)

		from ...general import SystematicAlternatives
		sa = SystematicAlternatives(
			masks=masks,
			groupby=groupby,
			categoricals=[
				u[0] if groupby_prefixes[n] is None else [groupby_prefixes[n]+str(uu) for uu in u[0]]
				for n,u in enumerate(uniqs)
			],
			altcodes=altcodes,
			padding_levels=padding_levels,
		)

		return x, sa

